diff --git a/src/libknot/packet/packet.c b/src/libknot/packet/packet.c
index 48a745f..51f4b8b 100644
--- a/src/libknot/packet/packet.c
+++ b/src/libknot/packet/packet.c
@@ -49,8 +49,6 @@ static int knot_packet_parse_question(knot_packet_t *pkt)
 	if (len <= 0)
 		return KNOT_EMALF;
 
-	/*! \todo Question case should be preserved. */
-	/* knot_dname_to_lower(question->qname); */
 	pkt->parsed += len + 2 * sizeof(uint16_t); /* Class + Type */
 	pkt->qname_size = len;
 
@@ -349,6 +347,7 @@ static void knot_packet_free_allocated_space(knot_packet_t *pkt)
 {
 	dbg_packet_verb("Freeing additional space in packet.\n");
 
+	pkt->mm.free(pkt->qname);
 	pkt->mm.free(pkt->answer);
 	pkt->mm.free(pkt->authority);
 	pkt->mm.free(pkt->additional);
@@ -739,8 +738,18 @@ const knot_dname_t *knot_packet_qname(const knot_packet_t *packet)
 	if (packet == NULL) {
 		return NULL;
 	}
+	if (packet->qname == NULL) {
+		/* Copy dname for lowercase conversion. */
+		knot_dname_t **qname = &((knot_packet_t *)packet)->qname;
+		*qname = packet->mm.alloc(packet->mm.ctx, packet->qname_size);
+		if (*qname == NULL) {
+			return NULL;
+		}
+		memcpy(*qname, packet->wireformat + KNOT_WIRE_HEADER_SIZE, packet->qname_size);
+		knot_dname_to_lower(*qname);
+	}
 
-	return packet->wireformat + KNOT_WIRE_HEADER_SIZE;
+	return packet->qname;
 }
 
 /*----------------------------------------------------------------------------*/
diff --git a/src/libknot/packet/packet.h b/src/libknot/packet/packet.h
index ad7e7b5..95d1534 100644
--- a/src/libknot/packet/packet.h
+++ b/src/libknot/packet/packet.h
@@ -91,6 +91,7 @@ struct knot_packet {
 	uint16_t parsed_ns;
 	uint16_t parsed_ar;
 
+	uint8_t *qname;
 	uint8_t qname_size; /*!< QNAME size. */
 	size_t size;      /*!< Current wire size of the packet. */
 	size_t max_size;  /*!< Maximum allowed size of the packet. */
